use {chumsky::span::SimpleSpan, lasso::Spur};

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct Spanned<T> {
    pub item: T,
    pub span: SimpleSpan,
}

impl<T> Spanned<T> {
    #[inline]
    pub const fn new(item: T, span: SimpleSpan) -> Self {
        Self { item, span }
    }

    pub fn map<U>(self, mut f: impl FnMut(T) -> U) -> Spanned<U> {
        Spanned {
            item: f(self.item),
            span: self.span,
        }
    }
}

pub type SpanExpr<'a> = Spanned<Expr<'a>>;
pub type ExprRef<'a> = &'a SpanExpr<'a>;
pub type ExprList<'a> = &'a [SpanExpr<'a>];
pub type Ident = Spur;

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum Expr<'a> {
    Ident(Ident),
    Path(ExprList<'a>),
    Literal(Literal),
    Call(ExprRef<'a>, ExprList<'a>),
    Binary(Spanned<BinaryOp>, ExprRef<'a>, ExprRef<'a>),
    Unary(Spanned<UnaryOp>, ExprRef<'a>),
    Set(ExprRef<'a>, ExprRef<'a>),
    Loop(ExprList<'a>),
    Block(ExprList<'a>),
    Definition(Definition<'a>),
    Switch {
        on:       ExprRef<'a>,
        branches: &'a [(Spanned<Switcher>, ExprRef<'a>)],
        else_:    ExprRef<'a>,
    },
    CondSwitch {
        branches: &'a [(ExprRef<'a>, ExprRef<'a>)],
        else_:    ExprRef<'a>,
    },
    Break(Option<ExprRef<'a>>),
    Return(Option<ExprRef<'a>>),
    Uninit,
    Continue,
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum Definition<'a> {
    Binding {
        kind:  DefKind,
        ident: Spanned<Ident>,
        ty:    Option<Spanned<Type>>,
        init:  ExprRef<'a>,
    },
    Func {
        name:   Spanned<Ident>,
        params: &'a [(Spanned<Ident>, Spanned<Type>)],
        ret:    Spanned<Type>,
        body:   ExprList<'a>,
    },
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum Type {
    Ident(Ident),
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum Switcher {
    Literal(Literal),
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum DefKind {
    Const,
    Var,
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum Literal {
    String(Spur),
    Int(u64),
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum BinaryOp {
    Plus,
    Minus,
    Star,
    Slash,
    Equ,
    Neq,
    Lt,
    Gt,
    LtEq,
    GtEq,
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum UnaryOp {
    Minus,
    Not,
    Star,
}
