use super::ast::{DefKind, Expr, ExprList, Type};

use {
    super::{
        ast::{Definition, Ident, SpanExpr, Spanned},
        token::{Token, T},
    },
    crate::utils::Pipe,
    bumpalo::Bump,
    chumsky::{
        extra::Full,
        input::{Stream, ValueInput},
        prelude::*,
    },
    logos::Lexer,
};

/// Equivalently-named unit variant mapping
macro_rules! equivmap {
    ($src:ident, $target:ident, [$variant0:ident $(, $variant:ident)* $(,)?] $(,)?) => {
        just($src::$variant0).to($target::$variant0)
            $(.or(just($src::$variant).to($target::$variant)))*
    };
}

fn ident<'a, I>() -> impl Parser<'a, I, Spanned<Ident>, Extra<'a>> + Clone + Copy
where
    I: Input<'a, Token = Token, Span = SimpleSpan> + ValueInput<'a>,
{
    select!(Token::Ident(id) => id).map_with_span(Spanned::new)
}

fn ty<'a, I>() -> impl Parser<'a, I, Spanned<Type>, Extra<'a>> + Clone + Copy
where
    I: Input<'a, Token = Token, Span = SimpleSpan> + ValueInput<'a>,
{
    ident().map(|i| i.map(Type::Ident))
}

fn definition<'a, I>() -> impl Parser<'a, I, Spanned<Definition<'a>>, Extra<'a>> + Clone
where
    I: Input<'a, Token = Token, Span = SimpleSpan> + ValueInput<'a>,
{
    let ident = ident();
    let ty = ty();

    let func = just(T!["func"])
        .ignore_then(ident)
        .then(
            ident
                .then_ignore(just(T![":"]))
                .then(ty)
                .separated_by(just(T![","]))
                .allow_trailing()
                .pipe(arena_collect)
                .delimited_by(just(T!["("]), just(T![")"])),
        )
        .then_ignore(just(T!["→"]))
        .then(ty)
        .then(just([T!["{"], T!["}"]]))
        .map_with_state(
            |(((name, params), ret), _body), _, state| Definition::Func {
                name,
                params,
                ret,
                body: state.arena.alloc_slice_copy(&[]),
            },
        );

    let binding = equivmap!(Token, DefKind, [Const, Var])
        .then(ident)
        .then(just(T![":"]).ignore_then(ty).or_not())
        .then(
            just(T!["="]).ignore_then(
                just(T!["uninit"])
                    .to(Expr::Uninit)
                    .map_with_span(Spanned::new),
            ),
        )
        .map_with_state(
            |(((kind, ident), ty), init), _, state| Definition::Binding {
                kind,
                ident,
                ty,
                init: state.arena.alloc(init),
            },
        );

    func.or(binding).map_with_span(Spanned::new)
}

pub struct State<'a> {
    pub arena: &'a Bump,
}

type Extra<'a> = Full<Rich<'a, Token>, State<'a>, ()>;
type ParseResult = ();

pub fn parse_input<'a>(
    input: impl ValueInput<'a, Token = Token, Span = SimpleSpan>,
    arena: &'a Bump,
) -> ParseResult {
    println!(
        "{:?}",
        definition()
            .separated_by(just(T![";"]))
            .allow_trailing()
            .pipe(arena_collect)
            .parse_with_state(input, &mut State { arena })
    );
}

pub fn parse_iter(
    input: impl Iterator<Item = (Token, SimpleSpan)>,
    eoi: impl Into<SimpleSpan>,
    arena: &Bump,
) -> ParseResult {
    parse_input(Stream::from_iter(input).spanned(eoi.into()), arena)
}

pub fn parse_lexer(input: Lexer<Token>, arena: &Bump) -> ParseResult {
    let end = input.span().end;
    parse_iter(
        input
            .spanned()
            .map(|(token, span)| (token.unwrap_or(Token::Invalid), span.into())),
        end..end + 1,
        arena,
    )
}

fn arena_collect<'a, I, O: 'a>(
    parser: impl IterParser<'a, I, O, Extra<'a>> + Clone,
) -> impl Parser<'a, I, &'a [O], Extra<'a>> + Clone
where
    I: Input<'a, Span = SimpleSpan, Token = Token>,
{
    empty()
        .map_with_state(|_, _, state: &mut State| bumpalo::vec![in state.arena])
        .foldl(parser, |mut v, o| {
            v.push(o);
            v
        })
        .map(bumpalo::collections::Vec::into_bump_slice)
}

fn arena_box<'a, I, O: 'a>(
    parser: impl Parser<'a, I, O, Extra<'a>> + Clone,
) -> impl Parser<'a, I, &'a O, Extra<'a>> + Clone
where
    I: Input<'a, Span = SimpleSpan, Token = Token>,
{
    parser.map_with_state(|item, _, state| &*state.arena.alloc(item))
}

#[inline]
fn merge_spans(start: SimpleSpan, end: SimpleSpan) -> SimpleSpan {
    SimpleSpan::new(start.start, end.end)
}
