use lasso::{Rodeo, Spur};
use logos::{Lexer, Logos};

#[derive(Default)]
pub struct Extras {
    pub interner: Rodeo,
}

macro_rules! token_def {
    (
        unit    { $($u_name:ident : $($u_tok:literal),* $(,)?;)* }
        keyword { $($kw:tt),* $(,)* }
        else    { $($e_tt:tt)* }
    ) => {
        literify::literify!(paste::paste! {
            #[derive(Clone, Copy, Debug, PartialEq, Eq, Logos)]
            #[logos(extras = Extras)]
            #[logos(skip r"[ \t\n\f]+")]
            #[logos(skip r"\\.*")]
            pub enum Token {
                $(
                    $(#[token($u_tok)])*
                    $u_name,
                )*

                $(
                    #[token(~($kw))]
                    [<$kw:camel>],
                )*

                $($e_tt)*
            }

            macro_rules! T {
                $($(
                    ($u_tok) => { $crate::syntax::token::Token::$u_name };
                )*)*

                $((~($kw)) => { $crate::syntax::token::Token::[<$kw:camel>] };)*
            }
        });
    };
}

token_def!(
    unit {
        LeftParen : "(";
        RightParen: ")";
        LeftCurly : "{";
        RightCurly: "}";
        Dot       : ".";
        Comma     : ",";
        Colon     : ":";
        Semicolon : ";";

        LArrow: "←", "<-";
        RArrow: "→", "->";

        Plus   : "+";
        Minus  : "-";
        Star   : "*";
        Slash  : "/";
        Precent: "%";

        Equ  : "=";
        Neq  : "≠", "/=";
        Lt   : "<";
        Gt   : ">";
        LtEq : "≤", "<=";
        GtEq : "≥", ">=";
    }

    keyword { func, var, const, include, switch, loop,
              return, break, continue, uninit, asm }

    else {
        #[regex(
            r"\p{XID_Start}\p{XID_Continue}*",
            |l| intern(l, l.slice()),
        )] Ident(Spur),

        #[regex(
            "\"[^\"]*\"",
            |l| {
                let s = l.slice();
                intern(l, &s[1..s.len() - 1])
            },
        )] String(Spur),

        #[regex(
            "[0-9]+",
            |l| l.slice().parse::<u64>().ok()
        )] Int(u64),

        Invalid,
    }
);

pub(crate) use T;

fn intern(lexer: &mut Lexer<'_, Token>, s: &str) -> Spur {
    lexer.extras.interner.get_or_intern(s)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn ident() {
        let mut lexer = Token::lexer("いえぶる able");
        assert_eq!(
            lexer.next(),
            Some(Ok(Token::Ident(intern(&mut lexer, "いえぶる")))),
        );

        assert_eq!(
            lexer.next(),
            Some(Ok(Token::Ident(intern(&mut lexer, "able")))),
        );
    }

    #[test]
    fn string() {
        let mut lexer = Token::lexer("\"sussy\" \"baka");
        assert_eq!(
            lexer.next(),
            Some(Ok(Token::String(intern(&mut lexer, "sussy")))),
        );

        assert_eq!(lexer.next(), Some(Err(())),);
    }

    #[test]
    fn symbol_alt() {
        let mut lexer = Token::lexer("-> →");
        assert_eq!(lexer.next(), lexer.next());
    }
}
